from PIL import Image
from skimage.metrics import structural_similarity as ssim
import numpy as np

from ._unify_shared import get_char, pre_render_characters, crop_image, get_font
from . import _unify_optim

WORKER_CHARACTERS = None

def set_worker_characters(characters):
    """Set the character list for the worker process."""
    global WORKER_CHARACTERS
    WORKER_CHARACTERS = characters

def compare_character(char, block_arr, save_chars, engine):
    if engine == "diff" or engine == "brightness":
        return _unify_optim.compare_character(char, block_arr, save_chars, engine)
    height, width = block_arr.shape
    
    char_arr = get_char(char, width, height, save=save_chars)
    if char_arr is None:
        return 0.0
    
    if char_arr.shape != block_arr.shape:
        char_img = Image.fromarray(char_arr)
        char_img = char_img.resize((width, height), Image.Resampling.BILINEAR)
        char_arr = np.array(char_img)
    
    # compare arrays
    if engine == "brightness":
        avg_char = np.mean(char_arr)
        avg_block = np.mean(block_arr)
        score = abs(avg_char - avg_block)
        similarity = 1.0 - (score / 255.0)
    elif engine == "ssim":
        min_side = min(block_arr.shape)
        win_size = min(7, min_side)
        if win_size % 2 == 0:
            win_size -= 1
        if win_size < 3:
            win_size = 3
        
        try:
            score = ssim(block_arr, char_arr, data_range=255, win_size=win_size)
            if isinstance(score, tuple):
                score = score[0]
            similarity = max(0.0, float(score))
        except ValueError:
            # This can happen if the window size is larger than the image
            similarity = 0.0
            
    elif engine == "diff":
        pixel_diff = np.sum(np.abs(block_arr.astype(np.int16) - char_arr.astype(np.int16)))
        max_diff = float(width * height * 255)
        if max_diff == 0:
            return 1.0
        dissimilarity = pixel_diff / max_diff
        similarity = 1.0 - dissimilarity
    elif engine == "mse":
        mse = np.mean((block_arr.astype(np.int32) - char_arr.astype(np.int32)) ** 2)
        max_mse = 255.0 ** 2
        similarity = 1.0 - (mse / max_mse) if max_mse > 0 else 1.0
    elif engine == "ncc":
        a = block_arr.astype(np.float64)
        b = char_arr.astype(np.float64)
        
        std_a = a.std()
        std_b = b.std()

        if std_a == 0 and std_b == 0:
            similarity = 1.0 if a.mean() == b.mean() else 0.0
        elif std_a == 0 or std_b == 0:
            similarity = 0.0
        else:
            a = (a - a.mean()) / std_a
            b = (b - b.mean()) / std_b
            ncc = np.mean(a * b)
            similarity = (ncc + 1) / 2
    elif engine == "hist":
        hist_a, _ = np.histogram(block_arr, bins=256, range=(0,255), density=True)
        hist_b, _ = np.histogram(char_arr, bins=256, range=(0,255), density=True)
        similarity = np.minimum(hist_a, hist_b).sum()
    elif engine == "cosine":
        a = block_arr.flatten().astype(np.float64)
        b = char_arr.flatten().astype(np.float64)
        
        norm_a = np.linalg.norm(a)
        norm_b = np.linalg.norm(b)

        if norm_a == 0 and norm_b == 0: # Both are black
            similarity = 1.0
        elif norm_a == 0 or norm_b == 0: # One is black
            similarity = 0.0
        else:
            cos_sim = np.dot(a, b) / (norm_a * norm_b)
            similarity = (cos_sim + 1) / 2
    else:
        raise ValueError(f"Invalid engine specified: {engine}")
        
    return similarity
    
def get_character(img_arr, engine, save_chars):
    """
    Finds the best character to represent an image block,
    using the globally defined WORKER_CHARACTERS list.
    """
    max_similarity = float("-inf")
    best_match = " "

    if WORKER_CHARACTERS is None:
        raise ValueError("Worker characters have not been initialized.")

    for character in WORKER_CHARACTERS:
        similarity = compare_character(character, img_arr, save_chars, engine)
        if similarity > max_similarity:
            max_similarity = similarity
            best_match = character
    
    return best_match

def process_blocks_batch(blocks_batch, engine, save_chars):
    """Processes a batch of blocks and returns a result map."""
    results_map = {}
    for block in blocks_batch:
        results_map[block.tobytes()] = get_character(block, engine, save_chars)
    return results_map




# TODO: video support

import sys
import time
import logging

import click
import itertools
import multiprocessing
import concurrent.futures
import collections
from PIL import Image, ImageSequence, UnidentifiedImageError
from moviepy import VideoFileClip
from tqdm import tqdm
from rich.console import Console
from rich.text import Text

from koba import __version__
from koba.core import core

try:
    multiprocessing.set_start_method("spawn")
except RuntimeError:
    pass

logging.basicConfig(
    format="{asctime} - {levelname} - {message}",
    style="{",
    datefmt="%Y-%m-%d %H:%M",
    level=logging.ERROR
)

@click.command("koba")
@click.version_option(__version__)
@click.argument(
    "file",
    type=click.Path(
        exists=True,
        file_okay=True,
        readable=True,
    )
)
@click.option(
    "--char-aspect", default=2.0, show_default=True,
    help="Character height-to-width ratio (for aspect-correct output)."
)
@click.option(
    "--logging-level",
    default="ERROR",
    show_default=True,
    help="Set the logging verbosity. Options: CRITICAL, ERROR, WARNING, INFO, DEBUG."
)
@click.option(
    "--save-blocks", is_flag=True,
    help="Save each image block as a PNG file in the 'blocks/' directory."
)
@click.option(
    "--save-chars", is_flag=True,
    help="Save rendered character images in the 'chars/' directory."
)
@click.option(
    "--engine", "-e", default="diff", show_default=True,
    help="Similarity metric to use: brightness, ssim, diff, mse, ncc, hist, or cosine."
)
@click.option(
    "--font", default=None,
    help="Path to a custom TTF font file (overrides the default font)."
)
@click.option(
    "--char-range", default="32-126", show_default=True,
    help="Unicode range of characters to use, as start-end (e.g., 32-126)."
)
@click.option(
    "--stretch-contrast",
    is_flag=True,
    help="Stretch image contrast to use the full brightness range."
)
@click.option(
    "--scale",
    default=1.0,
    help="Sets the scale at which to render the image. Will be overwritten if image needs to be scaled down to allow proper processing."
)
@click.option(
    "--invert",
    is_flag=True,
    help="Inverts the image for processing (Color will not be inverted when using --color)."
)
@click.option(
    "--single-threaded",
    is_flag=True,
    help="Runs the whole program single-threaded."
)
@click.option(
    "--color",
    is_flag=True,
    help="Renders the image in color."
)
@click.option(
    "--fast-color",
    is_flag=True,
    help="Enables color and uses â–ˆ (U+2588) to improve processing speed. Only recommended for animated pictures."
)
def main(file, char_aspect, logging_level, save_blocks, save_chars, engine, font, char_range, stretch_contrast, scale, invert, single_threaded, color, fast_color):
    console = Console()
    # update logging level
    logging.getLogger().setLevel(getattr(logging, logging_level.upper(), logging.ERROR))
    
    if fast_color:
        color = True
        char_range = "9608-9608"
    
    if font and not font.endswith(".ttf"):
        logging.critical("Please provide a font in TTF format.")
        sys.exit(1)

    split_range = char_range.split("-")
    start_char = split_range[0]
    end_char = split_range[1]
    if start_char.isnumeric() and end_char.isnumeric():
        start_char = int(start_char)
        end_char = int(end_char)
    else:
        logging.critical("Wrong format for char-range.")
        sys.exit(1)
    
    media_type = None
    
    # loading file and reading basic info
    try:
        img = Image.open(file)
        frame_count = getattr(img, 'n_frames', 1)
        if frame_count == 1:
            media_type = "image"
        else:
            media_type = "gif"
        frames = [frame.copy() for frame in ImageSequence.Iterator(img)]
    except UnidentifiedImageError:
        try:
            clip = VideoFileClip(file)
            frame_count = int(clip.fps * clip.duration)
            frames = (Image.fromarray(f) for f in clip.iter_frames())
            if frame_count > 1:
                media_type = "video"
            else:
                media_type = "image"
        except Exception as e:
            logging.critical(f"Unsupported or unreadable image/video format for file: {file}. Error: {e}")
            sys.exit(1)
    else:
        if not frames:
            logging.critical(f"No frames found in image file: {file}.")
            sys.exit(1)
    
    logging.info(f"Image has {frame_count} frame(s).")
    is_animated = frame_count > 1

    executor = None
    block_cache = collections.OrderedDict()
    characters = None

    try:
        if is_animated:
            from koba.core import unify, charsets
            logging.info("Pre-rendering characters...")
            if isinstance(frames, list):
                first_frame = frames[0]
            else:
                first_frame = next(frames)
                frames = itertools.chain([first_frame], frames)
            width, height = first_frame.size
            block_widths, block_heights, _ = core.calculate_block_sizes(width, height, char_aspect, scale)
            unique_shapes = {(w, h) for w in set(block_widths) for h in set(block_heights)}
            
            characters = charsets.get_range(start_char, end_char)
            
            if font:
                unify.font_path = font
            
            if abs(start_char - end_char) >= 400:
                for char in tqdm(characters, total=len(characters), desc="Loading fonts"):
                    unify.get_font(char)

            with tqdm(total=len(unique_shapes) * len(characters), desc="Pre-rendering characters", disable=logging_level != "DEBUG") as pbar:
                unify.pre_render_characters(characters, unique_shapes, save_chars, pbar.update)

            if not single_threaded:
                from koba.core.core import init_worker
                executor = concurrent.futures.ProcessPoolExecutor(initializer=init_worker, initargs=(characters,))

        frame_delays = []
        all_frames = []
        
        # Pass characters for single-threaded animated case
        chars_for_process = characters if (is_animated and single_threaded) else None

        for i, frame in tqdm(enumerate(frames), total=frame_count, desc="Processing frames", disable=not is_animated):
            all_frames.append(core.process(
                frame, char_aspect, scale, engine, color, invert, stretch_contrast,
                save_blocks, start_char, end_char, save_chars, font, 
                single_threaded=single_threaded, 
                show_progress=not is_animated,
                executor=executor,
                block_cache=block_cache,
                characters=chars_for_process
            ))
            delay = 0
            if media_type == "gif":
                delay = frame.info.get("duration", 100) / 1000
            elif media_type == "video":
                delay = 1 / clip.fps
            elif media_type == "image":
                delay = 0.1
            
            if not delay or delay == 0:
                delay = 0.1
            frame_delays.append(delay)

    finally:
        if executor:
            executor.shutdown()

    logging.debug(f"Frame delays: {frame_delays[:3]} ...")

    if media_type == "video":
        input("Press [Enter] to start playback: ")

    prev_lines = 0
    if is_animated:
        while True:
            for frame, delay in zip(all_frames, frame_delays):
                start = time.time()
                lines = frame.count('\n')
                if prev_lines > 0:
                    sys.stdout.write(f"\r\033[{prev_lines}A")
                    sys.stdout.write("\033[J")
                print(frame, end="")
                sys.stdout.flush()
                prev_lines = lines
                elapsed = time.time() - start
                sleep_time = max(0, delay - elapsed)
                time.sleep(sleep_time)
                
            if media_type == "video":
                try:
                    input("\nPress [Enter] to replay, or Ctrl+C to quit: ")
                    sys.stdout.write(f"\r\033[{prev_lines}A")
                    sys.stdout.write("\033[J")
                    sys.stdout.flush()
                except KeyboardInterrupt:
                    print("\nExiting.")
                    break
    else:
        frame = all_frames[0]
        if color:
            console.print(Text.from_ansi(frame))
        else:
            print(frame)


from setuptools import setup, Extension
from Cython.Build import cythonize
import numpy

extensions = [
    Extension(
        "koba.core._unify_optim",
        ["koba/core/_unify_optim.pyx"],
        include_dirs=[numpy.get_include()]
    )
]

setup(
    ext_modules=cythonize(extensions),
    zip_safe=False,
)



import os
import sys
import math
import logging
import concurrent.futures

import numpy as np
from tqdm import tqdm
from PIL import Image, ImageOps

from koba.core import charsets, unify

def init_worker(characters_for_worker):
    """Initializer for each worker process."""
    unify.set_worker_characters(characters_for_worker)

def chunk_list(data, n):
    """Splits a list into n roughly equal chunks."""
    if n <= 0:
        return [data]
    k, m = divmod(len(data), n)
    return [data[i*k+min(i, m):(i+1)*k+min(i+1, m)] for i in range(n)]

def calculate_block_sizes(width, height, char_aspect, scale):
    terminal_width = os.get_terminal_size().columns
    chars_width = terminal_width
    
    min_block_width = 10 / char_aspect
    min_block_height = 10

    scale = min(scale, 1.0)
    
    max_chars_width_by_block_width = int(width // min_block_width)
    max_chars_width_by_block_height = int(height // min_block_height * char_aspect)

    max_chars_width = min(max_chars_width_by_block_width, max_chars_width_by_block_height)

    chars_width = min(int(terminal_width * scale), terminal_width, max_chars_width)

    chars_width = max(chars_width, 1)

    chars_height = math.ceil((height * chars_width / width) / char_aspect)
    
    block_widths = [width // chars_width] * chars_width
    for i in range(width % chars_width):
        block_widths[i] += 1

    block_heights = [height // chars_height] * chars_height
    for i in range(height % chars_height):
        block_heights[i] += 1
        
    return block_widths, block_heights, chars_width

def process(
    img, char_aspect, scale, engine, color, invert, stretch_contrast, 
    save_blocks, start_char, end_char, save_chars, font, 
    single_threaded, show_progress=False, 
    executor=None, block_cache=None, characters=None
):
    if color:
        img_color = img.convert("RGB")
        img_arr_color = np.array(img_color)
    
    img = img.convert("L")
    if invert:
        img = ImageOps.invert(img)
    if stretch_contrast:
        img = ImageOps.autocontrast(img)
    
    img_arr = np.array(img)
    height, width = img_arr.shape[:2]
    logging.debug(f"Image is {width}x{height} pixels.")
    
    block_widths, block_heights, chars_width = calculate_block_sizes(width, height, char_aspect, scale)

    logging.debug(f"Image will be {chars_width}x{len(block_heights)} chars.")

    if min(block_heights + block_widths) <= 7 and str(engine).lower() == "ssim":
        logging.critical("Image blocks are too small for SSIM. Please use another engine.")
        sys.exit(1)

    # splitting into blocks with variable sizes
    block_colors = []
    blocks = []
    y = 0
    for bh in block_heights:
        x = 0
        for bw in block_widths:
            block = img_arr[y:y+bh, x:x+bw]                
            blocks.append(block)
            x += bw
        y += bh
    
    if color:
        y = 0
        for bh in block_heights:
            x = 0
            for bw in block_widths:
                block = img_arr_color[y:y+bh, x:x+bw]
                block_color = block.mean(axis=(0, 1)).astype(int)            
                block_colors.append(block_color)
                x += bw
            y += bh
    
    if save_blocks:
        os.makedirs("blocks", exist_ok=True)
        for idx, block in enumerate(blocks):
            Image.fromarray(block).save(os.path.join("blocks", f"{idx:04d}.png"))
    
    print_chars = ""
    
    if start_char == end_char:
        all_chars = chr(start_char) * len(blocks)
    else:
        frame_results_map = {}
        unique_blocks_map = {block.tobytes(): block for block in blocks}

        if block_cache is not None:
            blocks_to_process = []
            for key, block in unique_blocks_map.items():
                if key in block_cache:
                    frame_results_map[key] = block_cache[key]
                    block_cache.move_to_end(key)
                else:
                    blocks_to_process.append(block)
        else:
            blocks_to_process = list(unique_blocks_map.values())
        
        if blocks_to_process:
            new_results = {}
            if not single_threaded and executor:
                n_workers = executor._max_workers
                if not n_workers or n_workers <= 0: n_workers = 1
                block_chunks = chunk_list(blocks_to_process, n_workers)
                chunk_args = [(chunk, engine.lower(), save_chars) for chunk in block_chunks if chunk]
                futures = [executor.submit(unify.process_blocks_batch, *arg) for arg in chunk_args]
                for future in concurrent.futures.as_completed(futures):
                    new_results.update(future.result())
            else:
                if characters is None:
                    characters = charsets.get_range(start_char, end_char)
                init_worker(characters)
                for block in tqdm(blocks_to_process, desc="Processing unique blocks", disable=not show_progress):
                    new_results[block.tobytes()] = unify.get_character(block, engine.lower(), save_chars)

            if block_cache is not None:
                block_cache.update(new_results)
                while len(block_cache) > 1000000:
                    block_cache.popitem(last=False)
            frame_results_map.update(new_results)

        results = [frame_results_map.get(block.tobytes()) for block in blocks]
        all_chars = "".join(filter(None, results))
    
    lines = [all_chars[i:i+chars_width] for i in range(0, len(all_chars), chars_width)]

    if not color:
        return "\n".join(lines)

    # Apply colors
    print_chars = ""
    color_idx = 0
    for c in '\n'.join(lines):
        if c == '\n':
            print_chars += '\n'
        elif color_idx < len(block_colors):
            r, g, b = block_colors[color_idx]
            print_chars += f"\033[38;2;{r};{g};{b}m{c}\033[0m"
            color_idx += 1
        else:
            print_chars += c
            
    return print_chars


from PIL import ImageFont, ImageDraw, Image
import numpy as np
import logging
import os
from . import font

FONT_SIZE = 20
font_path = font.get_monospace_font()
font_cache = {}
char_cache = {}


def get_font(char):
    if char not in font_cache.keys():
        if font_path and font.check_support(char, font_path):
            font_cache[char] = ImageFont.truetype(font_path, FONT_SIZE)
        else:
            new_font = font.get_supported_font(char)
            if new_font:
                logging.debug(f"Found font that supports {char}.")
                font_cache[char] = ImageFont.truetype(new_font, FONT_SIZE)
            else:
                return None
    return font_cache[char]

# from UCYT5040/lectrick
def crop_image(image):
    image_array = np.array(image)
    non_white_mask = image_array < 0
    non_white_pixels = np.argwhere(non_white_mask)
    if non_white_pixels.size == 0:
        return image
    top, left = non_white_pixels.min(axis=0)
    bottom, right = non_white_pixels.max(axis=0) + 1  # +1 to include the last pixel
    return image.crop((left, top, right, bottom))

def get_char(char, width, height, save=False):
    char_arr = char_cache.get((char, width, height))
    if char_arr is not None:
        return char_arr
    else:
        font = get_font(char)
        if not font:
            logging.critical("Some chars are not supported by any of your installed fonts. Please specify a custom font with the --font option.")
            raise FileNotFoundError
        bbox = font.getbbox(char)
        
        if not (bbox and bbox[2] > bbox[0] and bbox[3] > bbox[1]):
            return None
        
        char_width, char_height = bbox[2] - bbox[0], bbox[3] - bbox[1]
        
        char_image = Image.new('L', (char_width, char_height), 0)
        draw = ImageDraw.Draw(char_image)
        draw.text((-bbox[0], -bbox[1]), char, font=font, fill=255)
                
        char_image = crop_image(char_image)
        
        char_image = char_image.resize((width, height), Image.Resampling.BILINEAR)
        
        if save:
            os.makedirs("chars", exist_ok=True)
            char_code = ord(char)
            char_image.save(f"chars/{char_code}.png")
        
        char_arr = np.array(char_image)
            
        char_cache[(char, width, height)] = char_arr
        return char_arr

def pre_render_characters(characters, sizes, save_chars, progress_callback=None):
    for size in sizes:
        width, height = size
        for char in characters:
            get_char(char, width, height, save_chars)
            if progress_callback:
                progress_callback()